#include        <stdio.h>

#include	"std.h"
#include	"code.h"


extern bool tflag;
extern bool vflag;
extern bool Vflag;

/*
##typespec cpu : unique
##*/
	void cpuProc ();

/*
##typespec cpu : unique
##*/
#define	GrNum	5

	typedef struct {
		char	*Cd;
		bool	Gr, Adr, Xr;
		void	(*Proc)();
	} CTAB;
	static void pld(), pst(), plea();
	static void padd(), psub(), pand(), por(), peor();
	static void pcpa(), pcpl(), psla(), psra(), psll(), psrl();
	static void pjpz(), pjmi(), pjnz(), pjze(), pjmp();
	static void ppush(), ppop(), pcall(), pret();
	static void pin(), pout(), poutint();
	static void pimul(), pidiv();
	static CTAB ctab[] = {
		{Null},
		{"LD",	True, True, True,	pld},
		{"ST",	True, True, True,	pst},
		{"LEA",	True, True, True,	plea},
		{"ADD",	True, True, True,	padd},
		{"SUB",	True, True, True,	psub},
		{"AND",	True, True, True,	pand},
		{"OR",	True, True, True,	por},
		{"EOR",	True, True, True,	peor},
		{"CPA",	True, True, True,	pcpa},
		{"CPL",	True, True, True,	pcpl},
		{"SLA",	True, True, True,	psla},
		{"SRA",	True, True, True,	psra},
		{"SLL",	True, True, True,	psll},
		{"SRL",	True, True, True,	psrl},
		{"JPZ",	False, True, True,	pjpz},
		{"JMI",	False, True, True,	pjmi},
		{"JNZ",	False, True, True,	pjnz},
		{"JZE",	False, True, True,	pjze},
		{"JMP",	False, True, True,	pjmp},
		{"PUSH", False, True, True,	ppush},
		{"POP",	True, False, False,	ppop},
		{"CALL", False, True, True,	pcall},
		{"RET",	False, False, False,	pret},
		{Null},
		{Null},
		{Null},
		{Null},
		{Null},
		{Null},
		{"IN",	False, True, False,	pin},
		{"OUT",	False, True, False,	pout},
		{Null},
		{"OUTINT", False, True, True,	poutint},
		{"MUL",	True, True, True,	pimul},
		{"DIV",	True, True, True,	pidiv},
		{Null},
		{Null},
		{Null},
		{Null},
		{Null},
		{Null},
		{Null},
	};

	static int pc;
	static int fr;
	static int grs[GrNum];

	static int gr, xr, adrs, adrs2;
	static int adx;

	static CTAB *ctabSrc ();
	static void frSet ();
	static int wdInt ();

	void cpuProc (staAdr) int staAdr; {
		CTAB *c;
		void (*proc)();
		int code, npc;

		errSet("run time error in address %04x", &pc);
		grs[4] = 10000;
		pc = atabGet(staAdr);
		for (;;) {
			npc = progGet(pc, &code, &gr, &adrs, &xr, &adrs2);
			if (code == CdExit) break;
			c = ctabSrc(code);
			if (!c) {
				pc = npc;
				continue;
			}
			if (c->Gr && !(0 <= gr && gr < GrNum)) {
				errPutD("illegal register specifier (%d)", gr);
				gr = 0;
			}
			adx = adrs;
			if (1 <= xr && xr < GrNum) {
				adx += grs[xr];
			}
			adx &= 0xffff;
			if ( Vflag ) {
			        printf("\t%04x %04x %04x %04x %04x %02x\n",
				       grs[0], grs[1], grs[2],
				               grs[3], grs[4], fr );
			}
			if (vflag) {
				printf("\t%04x\t%s\t", pc, c->Cd);
				if (c->Gr) printf("GR%d ", gr);
				if (c->Adr) printf("%04x ", adrs);
				if (c->Xr && 1 <= xr && xr <= 4)
					printf("GR%d ", xr);
				printf("\n");
			}
			if ( tflag )
			{
			        getc( stdin );
  		        }
			pc = npc;
			if (proc = c->Proc) (*proc)();
		}
	}

	static CTAB *ctabSrc (code) int code; {
		CTAB *c;

		if (code <= 0 || 40 <= code) {
			errPutD("unrecognized code (%d)", code);
			return (Null);
		}
		c = &ctab[code];
		if (!c->Cd) {
			errPutD("unrecognized code (%d)", code);
			return (Null);
		}
		return (c);
	}

	static void frSet (v) int v; {

		v &= 0xffff;
		if (v & 0x8000) fr = 0x10;
		else if (v == 0) fr = 0x01;
		else fr = 0x00;
	}

	static int wdInt (v) int v; {

		if (v & 0x8000) {
			v = -((-v) & 0xffff);
		}
		else {
			v &= 0xffff;
		}
		return (v);
	}


	static void pld () {

		grs[gr] = progLd(adx);
	}

	static void pst () {

		progSt(adx, grs[gr]);
	}

	static void plea () {

		grs[gr] = adx;
		frSet(adx);
	}

	static void padd () {

		grs[gr] = (grs[gr] + progLd(adx)) & 0xffff;
		frSet(grs[gr]);
	}

	static void psub () {

		grs[gr] = (grs[gr] - progLd(adx)) & 0xffff;
		frSet(grs[gr]);
	}

	static void pand () {

		grs[gr] = grs[gr] & progLd(adx);
		frSet(grs[gr]);
	}

	static void por () {

		grs[gr] = grs[gr] | progLd(adx);
		frSet(grs[gr]);
	}

	static void peor () {

		grs[gr] = grs[gr] ^ progLd(adx);
		frSet(grs[gr]);
	}

	static void pcpa () {
		int v;

		v = (grs[gr] - progLd(adx)) & 0xffff;
		frSet(v);
	}

	static void pcpl () {
		int v;
		int x,y;

/*		if (grs[gr]) v = 1;
		else v = 0;
		if (progLd(adx)) v--;	*/

/* Revised by K. Matsumoto on Jan 26 1993. */

		x = grs[gr] & 0xffff;
		y = progLd(adx) & 0xffff;
		if ( x > y ) v = 1;
		else if ( x < y ) v = -1;
		else v = 0;

		v &= 0xffff;
		frSet(v);
	}

	static void psla () {

		grs[gr] = (grs[gr] << adx) & 0xffff;
		frSet(grs[gr]);
	}

	static void psra () {

		grs[gr] = (wdInt(grs[gr]) >> adx) & 0xffff;
		frSet(grs[gr]);
	}

	static void psll () {

		grs[gr] = (grs[gr] << adx) & 0xffff;
		frSet(grs[gr]);
	}

	static void psrl () {

		grs[gr] = (grs[gr] >> adx) & 0xffff;
		frSet(grs[gr]);
	}

	static void pjpz () {

		if (fr == 0x00 || fr == 0x01) {
			pc = adx;
		}
	}

	static void pjmi () {

		if (fr == 0x10) {
			pc = adx;
		}
	}

	static void pjnz () {

		if (fr == 0x00 || fr == 0x10) {
			pc = adx;
		}
	}

	static void pjze () {

		if (fr == 0x01) {
			pc = adx;
		}
	}

	static void pjmp () {

		pc = adx;
	}

	static void ppush () {

		grs[4] -= 1;
		progSt(grs[4], adx);
	}

	static void ppop () {

		grs[gr] = progLd(grs[4]);
		grs[4] += 1;
	}

	static void pcall() {

		grs[4] -= 1;
		progSt(grs[4], pc);
		pc = adx;
	}

	static void pret() {

		pc = progLd(grs[4]);
		grs[4] += 1;
	}

	static void pin () {
	        int v, v2;

		for (v2 = 0; v2 < 80; ++v2, adrs ++ ) {
		        v = fgetc( stdin );
			if ( v == EOF || v == '\n' ) {
			        break;
			}
			progSt( adrs, v );
		}
		while ( v != '\n' && v != EOF ) {
		        v = fgetc( stdin );
		}
		if ( v == EOF && v2 == 0 ) {
		        v2 = -1;
		}
		progSt( adrs2, v2 );
	}

	static void pout () {
		int v, v2;

		v2 = progLd(adrs2);
		for (; v2 > 0; v2--, adrs++) {
			v = progLd(adrs);
			printf("%c", v & 0xff);
		}
		printf("\n");
	}

	static void poutint () {
		int v;

		v = progLd(adx);
		if (v & 0x8000) v = -(-v & 0xffff);
		printf("%d\n", v);
	}

	static void pimul () {

		grs[gr] = (wdInt(grs[gr]) * wdInt(progLd(adx))) & 0xffff;
		frSet(grs[gr]);
	}

	static void pidiv () {
		int v;

		v = wdInt(progLd(adx));
		if (v == 0) {
			errPutS("division by 0");
			exit(1);
		}
		grs[gr] = (wdInt(grs[gr]) / v) & 0xffff;
		frSet(grs[gr]);
	}

