

#include	"std.h"
#include	"code.h"


extern bool vflag;


/*
##typespec prog : unique
##*/
	void progInit ();
	void progReloc ();
	int  progGetLast ();
	void progPut ();
	void progPutS ();
	void progPutD ();
	int  progGet ();
	int  progLd ();
	void progSt ();

/*
##typespec atab : unique
##*/
	void atabInit ();
	void atabCheck ();
	int atabGet ();
	int atabSet ();
	int atabSetLabel ();
	int atabSetNum ();


/*
##typedef prog : unique
##*/
#define	SpaceSize	500000

	static int space[SpaceSize];
	static bool reloc[SpaceSize];
	static int sp;

	void progInit () {

		sp = 0;
	}

	void progReloc () {
		int adrs;

		for (adrs = 0; adrs < sp; adrs++) {
			if (reloc[adrs]) {
				space[adrs] = atabGet(space[adrs]);
				reloc[adrs] = False;
			}
		}
	}

	int progGetLast () {

		return (sp);
	}

	void progPut (code, gr, adr, xr, adr2) int code, gr, adr, xr, adr2; {

		if (code == CdOut || code == CdIn) {
			space[sp] = (code<<6);
			space[sp+1] = adr;	reloc[sp+1] = True;
			space[sp+2] = adr2;	reloc[sp+2] = True;
			sp += 3;
		}
		else {
			space[sp] = (code<<6) | ((gr&7)<<3) | (xr&7);
			space[sp+1] = adr;	reloc[sp+1] = True;
			sp += 2;
		}
	}

	void progPutS (size) int size; {

		sp += size;
	}

	void progPutD (adr) int adr; {

		space[sp] = adr;
		reloc[sp] = True;
		sp ++;
	}

	int progGet (adrs, code, gr, adr, xr, adr2)
		int adrs, *code, *gr, *adr, *xr, *adr2; {

		if (adrs < 0 || SpaceSize <= adrs) {
			errPutD("too high address (%d)", adrs);
			exit(1);
		}
		*code = (space[adrs]>>6);
		*gr = (space[adrs]>>3)&7;
		*xr = space[adrs]&7;
		if (*code == CdOut || *code == CdIn) {
			*adr = space[adrs+1];
			*adr2 = space[adrs+2];
			adrs += 3;
		}
		else {
			*adr = space[adrs+1];
			adrs += 2;
		}
		return (adrs);
	}

	int progLd (adrs) int adrs; {

		if (adrs < 0 || SpaceSize <= adrs) {
			errPutD("too high address (%d)", adrs);
			return (0);
		}
		return (space[adrs] & 0xffff);
	}

	void progSt (adrs, v) int adrs, v; {

		if (adrs < 0 || SpaceSize <= adrs) {
			errPutD("too high address (%d)", adrs);
			return;
		}
		space[adrs] = (v & 0xffff);
		reloc[adrs] = False;
	}

/*
##typedef atab : unique
##*/
#define	LabelLenMax	6

	typedef struct _ATAB {
		struct _ATAB	*Next;
		char	Label[LabelLenMax+1];
		int	Num;
		bool	Def;
	} ATAB;

	static ATAB *atab;

	static ATAB *atabPut ();
	static ATAB *atabSrcLabel ();
	static ATAB *atabSrcNum ();

	void atabInit () {

		atab = Null;
	}

	void atabCheck () {
		ATAB *p;
		int i;

		errSet("syntax error", &i);
		for (p = atab; p; p = p->Next) {
			if (!p->Def) {
				errPutS("undefined label (%s)", p->Label);
			}
			if (vflag && p->Label[0]) {
				printf("\t%s\t: %04x\n", p->Label, p->Num);
			}
		}
	}

	int atabGet (adr) int adr; {
		ATAB *p;

		if (!adr) return (0);
		p = (ATAB*)adr;
		return (p->Num);
	}

	int atabSet (label, num) char *label; int num; {
		ATAB *p;

		num &= 0xffff;
		p = atabSrcLabel(label);
		if (p) {
			if (p->Def) {
				errPutS("redefined label (%s)", label);
			}
			p->Num = num;  p->Def = True;
		}
		else {
			p = atabPut(label, num, True);
		}
		return (int)(p);
	}

	int atabSetLabel (label) char *label; {
		ATAB *p;

		p = atabSrcLabel(label);
		if (!p) {
			p = atabPut(label, 0, False);
		}
		return (int)(p);
	}

	int atabSetNum (num) int num; {
		ATAB *p;

		num &= 0xffff;
		p = atabSrcNum(num);
		if (!p) {
			p = atabPut("", num, True);
		}
		return (int)(p);
	}

	static ATAB *atabPut (label, num, def) char *label; int num; bool def; {
		ATAB *p;

		p = (ATAB*)malloc(sizeof(ATAB));
		strncpy(p->Label, label, LabelLenMax+1);
		p->Num = num;
		p->Def = def;
		p->Next = atab;
		atab = p;
		return (p);
	}

	static ATAB *atabSrcLabel (label) char *label; {
		ATAB *p;

		for (p = atab; p; p = p->Next) {
			if (!strcmp(p->Label, label)) return (p);
		}
		return (Null);
	}

	static ATAB *atabSrcNum (num) int num; {
		ATAB *p;

		for (p = atab; p; p = p->Next) {
			if (!p->Label[0] && p->Num == num) return (p);
		}
		return (Null);
	}


